/**
 * Window Options extension
 * v1.0.1
 * mathematical.coffee@gmail.com
 *
 * This extension adds 'minimize', 'maximize', ..., 'always on top'
 * etc options to the right-click menu of the title bar.
 *
 * It is meant to be compatible with the StatusTitleBar extension.
 *
 * FAQ:
 *   Q: How do I ensure this is loaded *after* Status Title Bar?
 *   A: Just turn it off and on a few times.
 *      (Interestingly if this is installed after Status Title Bar then it will
 *       always load after it).
 *
 * TODO:
 * - translations
 * - navigation options?
 * - workspace selector/show desktop?
 *
 */

const Lang = imports.lang;

const St = imports.gi.St;
const Meta = imports.gi.Meta;
const Wnck = imports.gi.Wnck;
const Shell = imports.gi.Shell;
const Clutter = imports.gi.Clutter;

const Main = imports.ui.main;
const Panel = imports.ui.panel;
const Tweener = imports.ui.tweener;
const PopupMenu = imports.ui.popupMenu;
const PanelMenu = imports.ui.panelMenu;
const Overview = imports.ui.overview;

const Util = imports.misc.util;


/* CONFIGURATION */

const config = {
    // if you want the more traditional
    // 'always on visible workspace'/'always on this workspace'
    // separate menu items or just one on/off toggle.
    workspace_toggle : true,
};

// Edit these if you want to change the text that appears
// (say you like 'Minimise' not 'Minimize', or 'Above' not 'Always on top').
const MINIMIZE                    = "Minimize";
const MAXIMIZE                    = "Maximize";
const RESTORE                     = "Restore";
const MOVE                        = "Move";
const RESIZE                      = "Resize";
const ALWAYS_ON_TOP               = "Always on top";
const ALWAYS_ON_VISIBLE_WORKSPACE = "Always on visible workspace";
const ALWAYS_ON_THIS_WORKSPACE    = "Always on this workspace";

// This is what will be in the menu.
// If you don't want an item, change it to 'false'.
// If you want a separator, make sure the string starts with
// 'separator' and hasn't already occurred in toAdd.
let toAdd = {};
toAdd[MINIMIZE] = true;
toAdd[RESTORE] = true;
toAdd[MAXIMIZE] = true;
toAdd['separator'] = true;

toAdd[MOVE] = true;
toAdd[RESIZE] = true;
toAdd['separator1'] = true;

toAdd[ALWAYS_ON_TOP] = true;
toAdd[ALWAYS_ON_VISIBLE_WORKSPACE] = true;
// Don't change this line.
toAdd[ALWAYS_ON_THIS_WORKSPACE] = !config.workspace_toggle;
toAdd['separator2'] = true;  // separates the above and 'Quit'.

/* CODE */

/*
 * WindowOptionsMenuHandler class.
 * Note: this adds to an *existing* menu.
 * It is not of itself a menu.
 */
function WindowOptionsMenuHandler() {
	this._init();
}

WindowOptionsMenuHandler.prototype = {
    _init: function () {
        this._items = {};
        this.theMenu = null;
        this._menuOpenedID = null;
        this._startupID = null;


        /* Try to get the 'Title Bar' object.
         * This is identified as the object with a ._targetApp property.
         */
        let children = Main.panel._leftBox.get_children();
        let i = children.length;
        while (i--) {
            if (children[i]._delegate &&
                    children[i]._delegate._targetApp !== undefined) {
                 // if _targetApp is set but NULL it will still pass this test.
                this.theMenu = children[i]._delegate;
                break;
            }
        }

        /* theMenu.menu is null until there is a focus app, so wait until then
         * to intialise the menu. Then we can stop listening to this event.
         */
        let tracker = Shell.WindowTracker.get_default();
        this._startupID = tracker.connect('notify::focus-app', Lang.bind(this, this._setupMenu));

    },

    _setupMenu: function () {
        let tracker = Shell.WindowTracker.get_default();
        tracker.disconnect(this._startupID);

        /* Add items */
        let posn = 0;
        for (let buttonName in toAdd) {
            if (toAdd.hasOwnProperty(buttonName) && toAdd[buttonName]) {
                if (buttonName.indexOf('separator') === 0) {
                    this._items[buttonName] = new PopupMenu.PopupSeparatorMenuItem();
                } else if (buttonName === ALWAYS_ON_TOP ||
                           (config.workspace_toggle && buttonName === ALWAYS_ON_VISIBLE_WORKSPACE)) {
                    // toggles
                    this._items[buttonName] = new PopupMenu.PopupSwitchMenuItem(buttonName, false);
                    this._items[buttonName].connect('toggled', Lang.bind(this, this._onActivate));
                } else {
                    this._items[buttonName] = new PopupMenu.PopupMenuItem(buttonName);
                    this._items[buttonName].connect('activate', Lang.bind(this, this._onActivate));
                }
                this.theMenu.menu.addMenuItem(this._items[buttonName], posn);
                posn++;
            }
        }

        /* Whenever menu is opened, update the toggles */
        /* Note: could listen to 'notify::above' but then I have to always track
         *  the current window & make sure it's connected, right?
         * In any case, I can't find a 'notify::on_all_workspaces' so that has
         * to use the menu-opened signal anyway - may as well do it for all.
         */
        if (this._items[ALWAYS_ON_TOP] ||
                (config.workspace_toggle && this._items[ALWAYS_ON_VISIBLE_WORKSPACE])) {
            this._menuOpenedID = this.theMenu.menu.connect('open-state-changed',
                                                           Lang.bind(this, this._updateToggles));
        }
    },

    destroy: function () {
        /* remove menu items */
        for (let buttonName in this._items) {
            if (this._items.hasOwnProperty(buttonName)) {
                this._items[buttonName].destroy();
            }
        }

        /* disconnect events */
        if (this._menuOpenedID) {
            this.theMenu.menu.disconnect(this._menuOpenedID);
        }
        if (this._startupID) {
            let tracker = Shell.WindowTracker.get_default();
            tracker.disconnect(this._startupID);
        }
    },

    _getCurrentWindowMutter: function () {
        let windows = this.theMenu._targetApp.get_windows();
        for (let i = 0; i < windows.length; i++) {
            if (windows[i].has_focus()) {
                return windows[i];
            }
        }
        // didn't find it.
        return null;
    },

    _getCurrentWindowWnck: function () {
        return Wnck.Screen.get_default().get_active_window();
    },

    _onActivate: function (clicked_actor) {
        let current_window = null;

        /* Ideally we want to use mutter for everything, but the JS
         * interface doesn't have move/resize/always on top/etc exposed,
         * so we have to use wnck until they do.
         */
        if (clicked_actor === this._items[MOVE]                            ||
                clicked_actor === this._items[RESIZE]                      ||
                clicked_actor === this._items[ALWAYS_ON_TOP]               ||
                clicked_actor === this._items[ALWAYS_ON_VISIBLE_WORKSPACE] ||
                clicked_actor === this._items[ALWAYS_ON_THIS_WORKSPACE]) {
            // go through wnck: not ideal, but these methods are not available through mutter yet.
            current_window = this._getCurrentWindowWnck();
        } else {
            current_window = this._getCurrentWindowMutter();
        }

        if (current_window) {
            // mutter methods
            if (clicked_actor === this._items[MINIMIZE]) {
                current_window.minimize();
            } else if (clicked_actor === this._items[MAXIMIZE]) {
                // NOTE: see Meta-1.0.gir for exported flags etc.
                current_window.maximize(Meta.MaximizeFlags.HORIZONTAL | Meta.MaximizeFlags.VERTICAL);
            } else if (clicked_actor === this._items[RESTORE]) {
                current_window.unmaximize(Meta.MaximizeFlags.HORIZONTAL | Meta.MaximizeFlags.VERTICAL);

            // The_following_are_wnck_methods
            } else if (clicked_actor === this._items[MOVE]) {
                current_window.keyboard_move();
            } else if (clicked_actor === this._items[RESIZE]) {
                current_window.keyboard_size();
            } else if (clicked_actor === this._items[ALWAYS_ON_TOP]) {
                if (current_window.is_above()) {
                    current_window.unmake_above();
                } else {
                    current_window.make_above();
                }
            /* Note: there is both a `pin` and a `stick` method.
             * I don't notice any difference between the two, but according to
             * http://developer.gnome.org/libwnck/stable/WnckWindow.html#wnck-window-is-pinned
             * `pinned` : "whether window is on all workspace. Pinned state may change anytime a
             *   'workspace-changed' signal gets emitted, but not when 'state-changed' gets emitted
             * `sticky` : "whether a window is sticky. May change any time a "state-changed" signal
             *   gets emitted.
             *
             * Sticky means "stuck to the glass", i.e. does not scroll with viewport.
             * In GDK/GTK+, sticky means "stuck to the glass" and *also* that the window is on all
             * workspaces. But here it only means the viewport aspect of it.
             *
             * Hence I think 'on all workspaces' is 'pin'. (But this can change with workspace-changed?!)
             */
            } else if (clicked_actor === this._items[ALWAYS_ON_VISIBLE_WORKSPACE]) {
                if (config.workspace_toggle && current_window.is_pinned()) {
                    current_window.unpin();
                } else {
                    current_window.pin();
                }
            } else if (clicked_actor === this._items[ALWAYS_ON_THIS_WORKSPACE]) {
                current_window.unpin();
            }
        }
    }, // _onActivate

    _updateToggles: function (obj, open) {
        /* Update the 'above'/'on all workspaces' toggles in case the user has
         * set this not using the menus.
         */
        if (!open) {
            return;
        }

        let current_window = this._getCurrentWindowMutter();
        if (this._items[ALWAYS_ON_TOP] &&
                (this._items[ALWAYS_ON_TOP].state !== current_window.above)) {
            this._items[ALWAYS_ON_TOP].setToggleState(current_window.above);
        }

        if (this._items[ALWAYS_ON_VISIBLE_WORKSPACE] &&
                (this._items[ALWAYS_ON_VISIBLE_WORKSPACE].state !== current_window.is_on_all_workspaces())) {
            this._items[ALWAYS_ON_VISIBLE_WORKSPACE].setToggleState(current_window.is_on_all_workspaces());
        }
    }
}; // end WindowOptionsMenuHandler.prototype

let windowOptions = null;

function init() {
}

function enable() {
    windowOptions = new WindowOptionsMenuHandler();
}

function disable() {
    if (windowOptions) {
        windowOptions.destroy();
    }
    windowOptions = null;
}

